@@ -1,5 +1,5 @@
 #
-# Copyright (C) 2014-2020 S[&]T, The Netherlands.
+# Copyright (C) 2014-2021 S[&]T, The Netherlands.
 #
 
 from __future__ import absolute_import, division, print_function
@@ -8,10 +8,13 @@ from muninn._compat import string_types as basestring
 import copy
 import datetime
 import errno
+import functools
+import hashlib
 import os
 import re
 import sys
 import uuid
+import warnings
 
 import muninn.config as config
 import muninn.util as util
@@ -23,6 +26,8 @@ from muninn.schema import *
 from muninn.struct import Struct
 from muninn import remote
 
+HASH_ALGORITHMS = set(hashlib.algorithms_guaranteed)
+
 
 class _ExtensionName(Text):
     _alias = "extension_name"
@@ -166,16 +171,38 @@ class Archive(object):
     def __exit__(self, type, value, traceback):
         self.close()
 
-    def _calculate_hash(self, product):
+    def _plugin_hash_type(self, plugin):
+        if hasattr(plugin, 'hash_type'):
+            if not plugin.hash_type:
+                return None
+            else:
+                return plugin.hash_type
+
+        elif hasattr(plugin, 'use_hash'):
+            if plugin.use_hash:
+                return 'sha1'
+            else:
+                return None
+
+        else:
+            return 'md5' # default after use_hash deprecation
+
+    def _extract_hash_type(self, hash_value):
+        prefix, middle, _ = hash_value.partition(':')
+        if middle == ':' and prefix in HASH_ALGORITHMS:
+            return prefix
+
+    def _calculate_hash(self, product, hash_type):
         """ calculate the hash on a product in the archive """
         product_path = self._product_path(product)
-        if not product_path:
+        if product_path is None:
             raise Error("no data available for product '%s' (%s)" % (product.core.product_name, product.core.uuid))
 
-        # Get the product type specific plug-in.
         plugin = self.product_type_plugin(product.core.product_type)
 
-        return self._storage.run_for_product(product, util.product_hash, plugin.use_enclosing_directory)
+        product_hash = functools.partial(util.product_hash, hash_type=hash_type)
+
+        return self._storage.run_for_product(product, product_hash, plugin.use_enclosing_directory)
 
     def _establish_invariants(self):
         repeat = True
@@ -217,7 +244,7 @@ class Archive(object):
                         self._purge(product)
 
     def _get_product(self, uuid):
-        products = self.search('uuid == @uuid', parameters={'uuid': uuid})
+        products = self.search('uuid == @uuid', parameters={'uuid': uuid}, namespaces=self.namespaces())
         if len(products) == 0:
             raise Error('No product found with UUID: %s' % uuid)
         assert len(products) == 1
@@ -230,18 +257,25 @@ class Archive(object):
         return self._storage.product_path(product)
 
     def _purge(self, product):
+        # get full product
+        product = self._get_product(product.core.uuid)
+
         # Remove the product from the product catalogue.
         self._database.delete_product_properties(product.core.uuid)
 
         # Remove any data in storage associated with the product.
         self._remove(product)
 
+        # Run the post remove hook (if defined by the product type plug-in).
+        plugin = self.product_type_plugin(product.core.product_type)
+        if hasattr(plugin, "post_remove_hook"):
+            plugin.post_remove_hook(self, product)
+
     def _relocate(self, product, properties=None):
         """Relocate a product to the archive_path reported by the product type plugin.
         Returns the new archive_path if the product was moved."""
         result = None
         product_archive_path = product.core.archive_path
-        product_path = self._product_path(product)
         if properties:
             product = copy.deepcopy(product)
             product.update(properties)
@@ -265,7 +299,7 @@ class Archive(object):
     def _retrieve(self, product, target_path, use_symlinks=False):
         # Determine the path of the product in storage.
         product_path = self._product_path(product)
-        if not product_path:
+        if product_path is None:
             raise Error("no data available for product '%s' (%s)" % (product.core.product_name, product.core.uuid))
 
         # Get the product type specific plug-in.
@@ -357,6 +391,17 @@ class Archive(object):
             self._database.delete_product_properties(product.core.uuid)
         return len(products)
 
+    def delete_properties_by_uuid(self, uuid):
+        """Remove properties from the catalogue for product with the given uuid.
+
+        This function will _not_ remove any product files from storage and will _not_ trigger any of the specific
+        cascade rules.
+
+        Keyword arguments:
+        uuid  --  UUID of the product whose metadata needs to be removed from the catalogue.
+        """
+        self._database.delete_product_properties(uuid)
+
     def derived_products(self, uuid):
         """Return the UUIDs of the products that are linked to the given product as derived products."""
         return self._database.derived_products(uuid)
@@ -405,9 +450,8 @@ class Archive(object):
             export_method_name = export_method_name + "_" + format
 
         result = []
-        products = self.search(where=where, parameters=parameters,
-                               property_names=['uuid', 'active', 'product_type', 'product_name', 'archive_path',
-                                               'physical_name'])
+        products = self.search(where=where, parameters=parameters, namespaces=self.namespaces())
+
         for product in products:
             if not product.core.active:
                 raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))
@@ -546,6 +590,13 @@ class Archive(object):
         # Extract product metadata.
         if properties is None:
             metadata = plugin.analyze(paths)
+            if hasattr(plugin, 'namespaces'):
+                namespaces = plugin.namespaces
+            else:
+                namespaces = []
+            for namespace in metadata:
+                if namespace != 'core' and namespace not in namespaces:
+                    warnings.warn("plugin.namespaces does not contain '%s'" % namespace, DeprecationWarning)
             if isinstance(metadata, (tuple, list)):
                 properties, tags = metadata
             else:
@@ -597,9 +648,10 @@ class Archive(object):
         try:
             # Determine product hash. Since it is an expensive operation, the hash is computed after inserting the
             # product properties so we won't needlessly compute it for products that fail ingestion into the catalogue.
-            if plugin.use_hash:
+            hash_type = self._plugin_hash_type(plugin)
+            if hash_type is not None:
                 try:
-                    properties.core.hash = util.product_hash(paths)
+                    properties.core.hash = util.product_hash(paths, hash_type=hash_type)
                 except EnvironmentError as _error:
                     raise Error("cannot determine product hash [%s]" % (_error,))
                 # Update the product hash in the product catalogue.
@@ -711,14 +763,15 @@ class Archive(object):
                 raise Error("more than one product found with name '%s'" % uuid_or_name_or_properties)
             product = products[0]
 
-        return self._product_path(product)
+        return os.path.join(self._storage.global_prefix, self._product_path(product))
 
     def product_type_plugin(self, product_type):
         """Return a reference to the product type plugin for a product type."""
         try:
             return self._product_type_plugins[product_type]
         except KeyError:
-            raise Error("undefined product type: \"%s\"; defined product types: %s" % (product_type, util.quoted_list(self._product_type_plugins.keys())))
+            raise Error("undefined product type: \"%s\"; defined product types: %s" %
+                        (product_type, util.quoted_list(self._product_type_plugins.keys())))
 
     def product_types(self):
         """Return a list of supported product types."""
@@ -738,7 +791,7 @@ class Archive(object):
                           the hash from the metadata (only if the metadata contained a hash).
 
         """
-        queue = self.search(where=where, parameters=parameters)
+        queue = self.search(where=where, parameters=parameters, namespaces=self.namespaces())
         for product in queue:
             if not product.core.active:
                 raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))
@@ -805,7 +858,7 @@ class Archive(object):
 
         # Determine the path of the product within storage
         product_path = self._product_path(product)
-        if not product_path:
+        if product_path is None:
             raise Error("no data available for product '%s' (%s)" % (product.core.product_name, product.core.uuid))
 
         # Extract product metadata.
@@ -836,6 +889,25 @@ class Archive(object):
             if new_archive_path:
                 properties.core.archive_path = new_archive_path
 
+        # if product type has disabled hashing, remove existing hash values
+        stored_hash = getattr(product.core, 'hash', None)
+        plugin_hash_type = self._plugin_hash_type(plugin)
+
+        if plugin_hash_type is None:
+            if stored_hash is not None:
+                properties.core.hash = None
+
+        # if product type has different hash algorithm, update hash values
+        else:
+            if stored_hash is None:
+                properties.core.hash = self._calculate_hash(product, plugin_hash_type)
+            else:
+                hash_type = self._extract_hash_type(stored_hash)
+                if hash_type is None and plugin_hash_type == 'sha1':
+                    properties.core.hash = plugin_hash_type + ':' + stored_hash
+                elif hash_type != plugin_hash_type:
+                    properties.core.hash = self._calculate_hash(product, plugin_hash_type)
+
         # Update product properties.
         self.update_properties(properties, uuid=product.core.uuid, create_namespaces=True)
 
@@ -921,9 +993,13 @@ class Archive(object):
             raise Error("redefinition of product type: \"%s\"" % product_type)
 
         # Quick verify of the plugin interface
-        for attr in ['use_enclosing_directory', 'use_hash']:
+        for attr in ['use_enclosing_directory']:
             if not hasattr(plugin, attr):
                 raise Error("missing '%s' attribute in plugin for product type \"%s\"" % (attr, product_type))
+
+        if hasattr(plugin, 'use_hash') and not hasattr(plugin, 'hash_type'):
+            warnings.warn("'use_hash' option is deprecated (use 'hash_type')")
+
         methods = ['identify', 'analyze', 'archive_path']
         if plugin.use_enclosing_directory:
             methods += ['enclosing_directory']
@@ -953,7 +1029,8 @@ class Archive(object):
         try:
             return self._remote_backend_plugins[remote_backend]
         except KeyError:
-            raise Error("undefined remote backend: \"%s\"; defined remote backends: %s" % (remote_backend, util.quoted_list(self._remote_backend.keys())))
+            raise Error("undefined remote backend: \"%s\"; defined remote backends: %s" %
+                        (remote_backend, util.quoted_list(self._remote_backend.keys())))
 
     def remote_backends(self):
         """Return a list of supported remote_backends."""
@@ -974,7 +1051,8 @@ class Archive(object):
                         this option with care.
         """
         products = self.search(where=where, parameters=parameters,
-                               property_names=['uuid', 'active', 'product_name', 'archive_path', 'physical_name', 'product_type'])
+                               property_names=['uuid', 'active', 'product_name', 'archive_path', 'physical_name',
+                                               'product_type'])
         for product in products:
             if not product.core.active and not force:
                 raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))
@@ -1021,7 +1099,7 @@ class Archive(object):
         return count
 
     def retrieve(self, where="", parameters={}, target_path=os.path.curdir, use_symlinks=False):
-        """Retrieve one or more products from the archive. Return the number of products retrieved.
+        """Retrieve one or more products from the archive. Return a list of target paths of the retrieved products.
 
         Keyword arguments:
         where           --  Search expression that determines which products to retrieve.
@@ -1032,6 +1110,7 @@ class Archive(object):
                             By default, products will be retrieved as copies.
 
         """
+        result = []
         products = self.search(where=where, parameters=parameters,
                                property_names=['uuid', 'active', 'product_type', 'product_name', 'archive_path',
                                                'physical_name'])
@@ -1039,9 +1118,9 @@ class Archive(object):
             if not product.core.active or 'archive_path' not in product.core:
                 raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))
 
-            self._retrieve(product, target_path, use_symlinks)
+            result.append(self._retrieve(product, target_path, use_symlinks))
 
-        return len(products)
+        return result
 
     def retrieve_by_name(self, product_name, target_path=os.path.curdir, use_symlinks=False):
         """Retrieve a product from the archive by name.
@@ -1056,26 +1135,26 @@ class Archive(object):
         An exception will be raised if no products with the specified name can be found.
 
         """
-        count = self.retrieve("product_name == @product_name", {"product_name": product_name}, target_path,
-                              use_symlinks)
-        if count == 0:
+        products = self.retrieve("product_name == @product_name", {"product_name": product_name}, target_path,
+                                 use_symlinks)
+        if len(products) == 0:
             raise Error("no products found with name '%s'" % product_name)
-        return count
+        return products
 
     def retrieve_by_uuid(self, uuid, target_path=os.path.curdir, use_symlinks=False):
         """Retrieve a product from the archive by uuid.
 
         This is a convenience function that is equivalent to:
 
-            self.retrieve("uuid == @uuid", {"uuid": uuid}, target_path, use_symlinks)
+            self.retrieve("uuid == @uuid", {"uuid": uuid}, target_path, use_symlinks)[0]
 
         An exception will be raised if no product with the specified uuid can be found.
 
         """
-        count = self.retrieve("uuid == @uuid", {"uuid": uuid}, target_path, use_symlinks)
-        if count == 0:
+        products = self.retrieve("uuid == @uuid", {"uuid": uuid}, target_path, use_symlinks)
+        if len(products) == 0:
             raise Error("product with uuid '%s' not found" % uuid)
-        return count
+        return products[0]
 
     def retrieve_properties(self, uuid, namespaces=[], property_names=[]):
         """Retrieve product properties for the product with the specified UUID.
@@ -1263,7 +1342,7 @@ class Archive(object):
                     raise Error("specified uuid does not match uuid included in the specified product properties")
             existing_product = self.search(where='uuid == @uuid', parameters={'uuid': uuid},
                                            namespaces=self.namespaces())[0]
-            new_namespaces = list(set(vars(properties).keys()) - set(vars(existing_product).keys()))
+            new_namespaces = list(set(vars(properties)) - set(vars(existing_product)))
         else:
             new_namespaces = None
         self._update_metadata_date(properties)
@@ -1291,6 +1370,17 @@ class Archive(object):
                 if 'hash' not in product.core:
                     raise Error("no hash available for product '%s' (%s)" %
                                 (product.core.product_name, product.core.uuid))
-                if self._calculate_hash(product) != product.core.hash:
+
+                stored_hash = product.core.hash
+                hash_type = self._extract_hash_type(stored_hash)
+
+                if hash_type is None:
+                    hash_type = 'sha1' # default before use_hash deprecation
+                    stored_hash = 'sha1:' + stored_hash
+
+                current_hash = self._calculate_hash(product, hash_type)
+
+                if current_hash != stored_hash:
                     failed_products.append(product.core.uuid)
+
         return failed_products
