@@ -1,5 +1,5 @@
 #
-# Copyright (C) 2014-2020 S[&]T, The Netherlands.
+# Copyright (C) 2014-2021 S[&]T, The Netherlands.
 #
 
 from __future__ import absolute_import, division, print_function
@@ -198,7 +198,6 @@ class PostgresqlConnection(object):
         else:
             raise Error('no such library: %s' % library)
 
-
     def __enter__(self):
         # Begin a transaction. The transaction is not started immediately, but a state change is recorded such that
         # attempts to start nested transactions can be detected. Also, the connection with the database is
@@ -238,7 +237,6 @@ class PostgresqlConnection(object):
         self._connection.close()
         self._connection = None
 
-
     def close(self):
         if self._in_transaction:
             raise InternalError("unable to close the connection with the database while a transaction is in progress")
@@ -288,12 +286,14 @@ class PostgresqlBackend(object):
         schema = self._namespace_schema("core")
         for name in schema:
             if schema.has_index(name):
-                result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
-                              (self._core_table_name, name, self._core_table_name, name))
+                if schema[name] == Geometry:
+                    # For the geospatial footprint we need to use a special GIST index
+                    result.append("CREATE INDEX idx_%s_%s ON %s USING GIST (%s)" %
+                                  (self._core_table_name, name, self._core_table_name, name))
+                else:
+                    result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
+                                  (self._core_table_name, name, self._core_table_name, name))
 
-        # For the geospatial footprint we need to use a special GIST index
-        result.append("CREATE INDEX idx_%s_footprint ON %s USING GIST (footprint)" %
-                      (self._core_table_name, self._core_table_name))
 
         # Create the tables for all non-core namespaces.
         for namespace in self._namespace_schemas:
@@ -310,8 +310,12 @@ class PostgresqlBackend(object):
             schema = self._namespace_schema(namespace)
             for name in schema:
                 if schema.has_index(name):
-                    result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
-                                  (self._table_name(namespace), name, self._table_name(namespace), name))
+                    if schema[name] == Geometry:
+                        result.append("CREATE INDEX idx_%s_%s ON %s USING GIST (%s)" %
+                                      (self._table_name(namespace), name, self._table_name(namespace), name))
+                    else:
+                        result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
+                                      (self._table_name(namespace), name, self._table_name(namespace), name))
 
         # We use explicit 'id' primary keys for the links and tags tables so the entries can be managed using
         # other front-ends that may not support tuples as primary keys.
@@ -458,6 +462,14 @@ class PostgresqlBackend(object):
         finally:
             cursor.close()
 
+    def _delete_namespace_properties(self, uuid, name):
+        query = "DELETE FROM %s WHERE uuid=%s" % (self._table_name(name), self._placeholder())
+        cursor = self._connection.cursor()
+        try:
+            cursor.execute(query, [uuid])
+        finally:
+            cursor.close()
+
     def _swallow_unique_violation(self, _error):
         # There is still a small chance due to concurrency that a link/tag already exists.
         # For those cases we swallow the exception.
@@ -470,7 +482,7 @@ class PostgresqlBackend(object):
             except AttributeError:
                 pass
 
-        elif self._library == 'pg8000': # TODO positional - issue filed on github
+        elif self._library == 'pg8000':  # TODO positional - issue filed on github
             try:
                 if _error.args[0]['C'] == PG_UNIQUE_VIOLATION:
                     swallow = True
@@ -591,6 +603,7 @@ class PostgresqlBackend(object):
         rewriter_table[Prototype("is_defined", (Real,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Boolean,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Text,), Boolean)] = is_defined_rewriter
+        rewriter_table[Prototype("is_defined", (Namespace,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Timestamp,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (UUID,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Geometry,), Boolean)] = is_defined_rewriter
@@ -657,8 +670,8 @@ class PostgresqlBackend(object):
             parameters = (uuid,)
         else:
             for source_uuid in source_uuids:
-                query = "DELETE FROM %s WHERE uuid = %s AND source_uuid = %s" % (self._link_table_name,
-                                                                                 self._placeholder(), self._placeholder())
+                query = "DELETE FROM %s WHERE uuid = %s AND source_uuid = %s" % \
+                    (self._link_table_name, self._placeholder(), self._placeholder())
                 parameters = (uuid, source_uuid)
                 cursor = self._connection.cursor()
                 try:
@@ -944,5 +957,7 @@ class PostgresqlBackend(object):
                     continue
                 if ns_name in new_namespaces:
                     self._insert_namespace_properties(uuid, ns_name, ns_properties)
+                elif ns_properties is None:
+                    self._delete_namespace_properties(uuid, ns_name)
                 else:
                     self._update_namespace_properties(uuid, ns_name, ns_properties)
