@@ -1,5 +1,5 @@
 #
-# Copyright (C) 2014-2020 S[&]T, The Netherlands.
+# Copyright (C) 2014-2021 S[&]T, The Netherlands.
 #
 
 from __future__ import absolute_import, division, print_function
@@ -255,11 +255,13 @@ class SQLiteBackend(object):
                               (self._core_table_name, name,))
         for name in schema:
             if schema.has_index(name):
-                result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
-                              (self._core_table_name, name, self._core_table_name, name))
+                if schema[name] == Geometry:
+                    # For the geospatial footprint we need to use a special spatial index
+                    result.append("SELECT CreateSpatialIndex('%s', '%s')" % (self._core_table_name, name))
+                else:
+                    result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
+                                  (self._core_table_name, name, self._core_table_name, name))
 
-        # For the geospatial footprint we need to use a special spatial index
-        result.append("SELECT CreateSpatialIndex('%s', '%s')" % (self._core_table_name, 'footprint'))
 
         # Create the tables for all non-core namespaces.
         for namespace in self._namespace_schemas:
@@ -284,8 +286,11 @@ class SQLiteBackend(object):
                                   (self._table_name(namespace), name))
             for name in schema:
                 if schema.has_index(name):
-                    result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
-                                  (self._table_name(namespace), name, self._table_name(namespace), name))
+                    if schema[name] == Geometry:
+                        result.append("SELECT CreateSpatialIndex('%s', '%s')" % (self._table_name(namespace), name))
+                    else:
+                        result.append("CREATE INDEX idx_%s_%s ON %s (%s)" %
+                                      (self._table_name(namespace), name, self._table_name(namespace), name))
 
         # We use explicit 'id' primary keys for the links and tags tables so the entries can be managed using
         # other front-ends that may not support tuples as primary keys.
@@ -440,6 +445,14 @@ class SQLiteBackend(object):
         finally:
             cursor.close()
 
+    def _delete_namespace_properties(self, uuid, name):
+        query = "DELETE FROM %s WHERE uuid=%s" % (self._table_name(name), self._placeholder())
+        cursor = self._connection.cursor()
+        try:
+            cursor.execute(query, [uuid])
+        finally:
+            cursor.close()
+
     def _link(self, uuid, source_uuids):
         query = "INSERT OR IGNORE INTO %s (uuid, source_uuid) VALUES (%s, %s)" % \
             (self._link_table_name, self._placeholder(), self._placeholder())
@@ -552,6 +565,7 @@ class SQLiteBackend(object):
         rewriter_table[Prototype("is_defined", (Real,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Boolean,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Text,), Boolean)] = is_defined_rewriter
+        rewriter_table[Prototype("is_defined", (Namespace,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Timestamp,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (UUID,), Boolean)] = is_defined_rewriter
         rewriter_table[Prototype("is_defined", (Geometry,), Boolean)] = is_defined_rewriter
@@ -884,6 +898,8 @@ class SQLiteBackend(object):
                     continue
                 if ns_name in new_namespaces:
                     self._insert_namespace_properties(uuid, ns_name, ns_properties)
+                elif ns_properties is None:
+                    self._delete_namespace_properties(uuid, ns_name)
                 else:
                     self._update_namespace_properties(uuid, ns_name, ns_properties)
 
