@@ -1,5 +1,5 @@
 #
-# Copyright (C) 2014-2020 S[&]T, The Netherlands.
+# Copyright (C) 2014-2021 S[&]T, The Netherlands.
 #
 
 from __future__ import absolute_import, division, print_function
@@ -20,19 +20,28 @@ from muninn.schema import *
 from muninn.visitor import Visitor
 
 #
-# Table of all supported operators and functions.
+# Table of all supported operators and functions
 #
 function_table = FunctionTable()
-
 #
-# Logical operators.
+# Logical operators
 #
 function_table.add(Prototype("not", (Boolean,), Boolean))
 function_table.add(Prototype("and", (Boolean, Boolean), Boolean))
 function_table.add(Prototype("or", (Boolean, Boolean), Boolean))
-
 #
-# Comparison operators.
+# Membership operators
+#
+function_table.add(Prototype("in", (Integer, Sequence), Boolean))
+function_table.add(Prototype("in", (Long, Sequence), Boolean))
+function_table.add(Prototype("in", (Real, Sequence), Boolean))
+function_table.add(Prototype("in", (Text, Sequence), Boolean))
+function_table.add(Prototype("not in", (Integer, Sequence), Boolean))
+function_table.add(Prototype("not in", (Long, Sequence), Boolean))
+function_table.add(Prototype("not in", (Real, Sequence), Boolean))
+function_table.add(Prototype("not in", (Text, Sequence), Boolean))
+#
+# Comparison operators
 #
 function_table.add(Prototype("==", (Long, Long), Boolean))
 function_table.add(Prototype("==", (Long, Integer), Boolean))
@@ -174,6 +183,7 @@ function_table.add(Prototype("is_defined", (Integer,), Boolean))
 function_table.add(Prototype("is_defined", (Real,), Boolean))
 function_table.add(Prototype("is_defined", (Boolean,), Boolean))
 function_table.add(Prototype("is_defined", (Text,), Boolean))
+function_table.add(Prototype("is_defined", (Namespace,), Boolean))
 function_table.add(Prototype("is_defined", (Timestamp,), Boolean))
 function_table.add(Prototype("is_defined", (UUID,), Boolean))
 function_table.add(Prototype("is_defined", (Geometry,), Boolean))
@@ -206,9 +216,8 @@ class TokenStream(object):
             r"""[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}""",  # UUID literals
             r"""\d+(?:\.\d*(?:[eE][+-]?\d+)?|[eE][+-]?\d+)""",               # Real literals
             r"""\d+""",                                                      # Integer literals
-            r"""true|false""",                                               # Boolean literals
-            r"""[a-zA-Z]\w*""",                                              # Names
-            r"""<=|>=|==|!=|~=|[*<>@(),.+-/]"""                              # Operators and delimiters
+            r"""<=|>=|==|!=|~=|not in|[*<>@()\[\],.+-/]""",                  # Operators and delimiters
+            r"""[a-zA-Z]\w*""",                                              # Names (incl. true, false, in)
         )
 
     _pattern = r"""(?:%s)""" % ("|".join(["(%s)" % sub_pattern for sub_pattern in _sub_patterns]))
@@ -304,32 +313,34 @@ class TokenStream(object):
                         self.text[self.token_start_position:]))
 
         self.token_start_position, self.token_end_position = match_object.span()
-        text, timestamp, uuid_, real, integer, boolean, name, operator = match_object.groups()
+        text, timestamp, uuid_, real, integer, operator, name = match_object.groups()
 
         if text is not None:
             return Token(TokenType.TEXT, string_unescape(text[1:-1]))
 
-        if uuid_ is not None:
-            return Token(TokenType.UUID, uuid.UUID(uuid_))
-
         if timestamp is not None:
             return Token(TokenType.TIMESTAMP, self._parse_timestamp(timestamp))
 
+        if uuid_ is not None:
+            return Token(TokenType.UUID, uuid.UUID(uuid_))
+
         if real is not None:
             return Token(TokenType.REAL, float(real))
 
         if integer is not None:
             return Token(TokenType.INTEGER, int(integer))
 
-        if boolean is not None:
-            return Token(TokenType.BOOLEAN, boolean == "true")
-
-        if name is not None:
-            return Token(TokenType.NAME, name)
-
         if operator is not None:
             return Token(TokenType.OPERATOR, operator)
 
+        if name is not None:
+            if name in ["true", "false"]:
+                return Token(TokenType.BOOLEAN, name == "true")
+            elif name == "in":
+                return Token(TokenType.OPERATOR, name)
+            else:
+                return Token(TokenType.NAME, name)
+
         raise Error("char %d: syntax error: \"%s\"" % (self.token_start_position + 1, match_object.group()))
 
     def _skip_white_space(self, start):
@@ -372,6 +383,13 @@ class Name(AbstractSyntaxTreeNode):
     def __str__(self):
         return "(%s %s)" % (type(self).__name__, self.value)
 
+class List(AbstractSyntaxTreeNode):
+    def __init__(self, values):
+        self.values = values
+
+    def __str__(self):
+        return "(%s %s)" % (type(self).__name__, self.value)
+
 
 class ParameterReference(AbstractSyntaxTreeNode):
     def __init__(self, name):
@@ -392,15 +410,15 @@ class FunctionCall(AbstractSyntaxTreeNode):
         return "(%s %s %s)" % (type(self).__name__, self.name, " ".join(map(str, self.arguments)))
 
 
-def parse_sequence(stream, parse_item_function):
-    stream.expect(TokenType.OPERATOR, "(")
-    if stream.accept(TokenType.OPERATOR, ")"):
+def parse_sequence(stream, parse_item_function, start='(', end=')'):
+    stream.expect(TokenType.OPERATOR, start)
+    if stream.accept(TokenType.OPERATOR, end):
         return []
 
     sequence = [parse_item_function(stream)]
     while stream.accept(TokenType.OPERATOR, ","):
         sequence.append(parse_item_function(stream))
-    stream.expect(TokenType.OPERATOR, ")")
+    stream.expect(TokenType.OPERATOR, end)
     return sequence
 
 
@@ -513,6 +531,9 @@ def parse_atom(stream):
             parts.append(name_token.value)
         return Name(".".join(parts))
 
+    if stream.test(TokenType.OPERATOR, "["):
+        return List(parse_sequence(stream, parse_expression, "[", "]"))
+
     # Literal.
     token = stream.expect((TokenType.TEXT, TokenType.TIMESTAMP, TokenType.UUID, TokenType.REAL, TokenType.INTEGER,
                            TokenType.BOOLEAN))
@@ -536,8 +557,8 @@ def parse_arithmetic_expression(stream):
 
 def parse_comparison(stream):
     lhs = parse_arithmetic_expression(stream)
-    if stream.test(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=")):
-        operator_token = stream.expect(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~="))
+    if stream.test(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in")):
+        operator_token = stream.expect(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in"))
         return FunctionCall(operator_token.value, lhs, parse_comparison(stream))
     return lhs
 
@@ -614,17 +635,29 @@ class SemanticAnalysis(Visitor):
         # namespace
         if name is None:
             visitable.value = split_name[0]
-            visitable.type = Text
+            visitable.type = Namespace
 
         # namespace.property
         else:
             try:
-                type = schema[name]
+                type_ = schema[name]
             except KeyError:
-                raise Error("no property: \"%s\" defined within namespace: \"%s\"" % (name, namespace))
+                if len(split_name) == 2:
+                    raise Error("undefined property: \"%s\"" % visitable.value)
+                else:
+                    raise Error("undefined name: \"%s\"" % name)
 
             visitable.value = "%s.%s" % (namespace, name)
-            visitable.type = type
+            visitable.type = type_
+
+    def visit_List(self, visitable): # TODO check same literal type
+        values = []
+        for value in visitable.values:
+            if not isinstance(value, Literal):
+                raise Error("list contains non-literal")
+            values.append(value.value)
+        visitable.value = values
+        visitable.type = Sequence
 
     def visit_ParameterReference(self, visitable):
         try:
