@@ -32,7 +32,7 @@ os.environ['MUNINN_CONFIG_PATH'] = '.'
 
 import shutil
 
-from muninn.schema import Mapping, optional, Text
+from muninn.schema import Mapping, optional, Text, Integer
 from muninn.geometry import Polygon, LinearRing, Point
 
 CFG = ConfigParser()
@@ -47,6 +47,9 @@ USE_ENCLOSING_DIR = [s.strip()=='true' for s in CFG.get('DEFAULT', 'use_enclosin
 class MyNamespace(Mapping):
     hello = optional(Text)
 
+class MyNamespace2(Mapping):
+    counter = optional(Integer)
+
 
 class BaseChecker(object):
     def __init__(self, storage):
@@ -180,6 +183,7 @@ def archive(database, storage, use_enclosing_directory, archive_path):
     # create clean archive
     with muninn.open('my_arch') as archive:
         archive.register_namespace('mynamespace', MyNamespace)
+        archive.register_namespace('mynamespace2', MyNamespace2)
         archive.destroy()
         archive.prepare()
 
@@ -217,7 +221,6 @@ class TestArchive:
             verify_hash=True,
             use_symlinks=use_symlinks
         )
-
         path = os.path.join(archive._params['archive_path'], 'pi.txt')
 
         if archive._params['use_enclosing_directory']:
@@ -676,6 +679,56 @@ class TestArchive:
                 )
                 assert not archive._checker.exists(oldpath)
 
+    def test_rebuild_properties_hash(self, archive):
+        plugin = archive._product_type_plugins['MY_TYPE']
+
+        try:
+            # ingest product: hash_type is set to 'md5'
+            product = self._ingest_file(archive)
+            assert product.core.hash.startswith('md5:')
+
+            # plugin.hash_type changed, so hash should change on rebuild_properties
+            archive._product_type_plugins['MY_TYPE'].hash_type = 'sha1'
+            archive.rebuild_properties(product.core.uuid)
+            product = archive.retrieve_properties(product.core.uuid)
+            assert product.core.hash.startswith('sha1:')
+
+            # remove stored hash, so rebuild_properties should rebuild it
+            product.core.hash = None
+            archive.update_properties(product)
+            archive.rebuild_properties(product.core.uuid)
+            product = archive.retrieve_properties(product.core.uuid)
+            assert product.core.hash.startswith('sha1:')
+
+            # upgrade case: stored hash without prefix, so add it
+            product.core.hash = product.core.hash[len('sha1:'):]
+            archive.update_properties(product)
+            archive.rebuild_properties(product.core.uuid)
+            product = archive.retrieve_properties(product.core.uuid)
+            assert product.core.hash.startswith('sha1:')
+
+            # hashing was disabled, so remove hash
+            archive._product_type_plugins['MY_TYPE'].hash_type = None
+            archive.rebuild_properties(product.core.uuid)
+            product = archive.retrieve_properties(product.core.uuid)
+            assert 'hash' not in product.core
+
+        finally:
+            plugin.hash_type = 'md5'
+
+    def test_verify_hash(self, archive):
+        product = self._ingest_file(archive)
+
+        # normal case: hash contains prefix
+        failed = archive.verify_hash()
+        assert len(failed) == 0
+
+        # upgrade case: no hash prefix, so fail on md5 vs sha1
+        product.core.hash = product.core.hash[len('md5:'):]
+        archive.update_properties(product)
+        failed = archive.verify_hash()
+        assert len(failed) == 1
+
     def test_rebuild_pull_properties(self, archive):
         properties = self._pull(archive)
         archive.rebuild_pull_properties(properties.core.uuid, verify_hash=True)
@@ -833,6 +886,68 @@ class TestQuery:
         s = archive.search('is_defined(physical_name)')
         assert len(s) == 3
 
+    def test_InList(self, archive):
+        self._prep_data(archive)
+        s = archive.search('size in [1015]')
+        assert len(s) == 3
+        s = archive.search('not size in [1015]')
+        assert len(s) == 0
+        s = archive.search('size not in [1015]')
+        assert len(s) == 0
+
+        s = archive.search('size in [1016]')
+        assert len(s) == 0
+        s = archive.search('not size in [1016]')
+        assert len(s) == 3
+        s = archive.search('size not in [1016]')
+        assert len(s) == 3
+
+        s = archive.search('size in [1015, 1016]')
+        assert len(s) == 3
+        s = archive.search('not size in [1015, 1016]')
+        assert len(s) == 0
+        s = archive.search('size not in [1015, 1016]')
+        assert len(s) == 0
+
+        s = archive.search('size in [1016, 1017]')
+        assert len(s) == 0
+        s = archive.search('not size in [1016, 1017]')
+        assert len(s) == 3
+        s = archive.search('size not in [1016, 1017]')
+        assert len(s) == 3
+
+        s = archive.search('physical_name in ["a.txt"]')
+        assert len(s) == 1
+        s = archive.search('not physical_name in ["a.txt"]')
+        assert len(s) == 2
+        s = archive.search('physical_name not in ["a.txt"]')
+        assert len(s) == 2
+
+        s = archive.search('physical_name in ["b.txt", "c.txt"]')
+        assert len(s) == 2
+        s = archive.search('not physical_name in ["b.txt", "c.txt"]')
+        assert len(s) == 1
+        s = archive.search('physical_name not in ["b.txt", "c.txt"]')
+        assert len(s) == 1
+
+    def test_RemoveProperties(self, archive):
+        self._prep_data(archive)
+
+        p = archive.retrieve_properties(uuid=self.uuid_a, namespaces=['mynamespace'])
+        assert hasattr(p, 'core')
+        assert hasattr(p, 'mynamespace')
+        assert hasattr(p.mynamespace, 'hello')
+
+        # remove property
+        archive.update_properties(muninn.Struct({'mynamespace': muninn.Struct({'hello': None})}), p.core.uuid)
+        p = archive.retrieve_properties(uuid=self.uuid_a, namespaces=['mynamespace'])
+        assert not hasattr(p.mynamespace, 'hello')
+
+        # remove namespace
+        archive.update_properties(muninn.Struct({'mynamespace': None}), p.core.uuid)
+        p = archive.retrieve_properties(uuid=self.uuid_a, namespaces=['mynamespace'])
+        assert not hasattr(p, 'mynamespace')
+
     def test_Geometry(self, archive):
         self._prep_data(archive)
 
